# -*- coding: utf-8 -*-
# Copyright 2014-17 Neil Freeman
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from datetime import datetime
from pytz import timezone

ISOFORMAT = '%Y-%m-%dT%H:%M:%S'


def parse_datetime(dt):
    '''Parse an ISO datetime, which Python does buggily.'''
    d = datetime.strptime(dt[:-1], ISOFORMAT)

    if dt[-1:] == 'Z':
        return timezone('utc').localize(d)
    else:
        return d


def iso_format(dt):
    return dt.strftime(ISOFORMAT + 'Z')



# -*- coding: utf-8 -*-
# Copyright 2014-17 Neil Freeman
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import csv
import re
from io import BytesIO, StringIO
import requests
from pytz import utc
from . import properties, timezone

# Both take station as a GET argument.
OBS_ENDPOINT = "http://sdf.ndbc.noaa.gov/sos/server.php"
CAM_ENDPOINT = 'http://www.ndbc.noaa.gov/buoycam.php'

'''
request=GetObservation
service=SOS
version=1.0.0
offering=urn:ioos:station:wmo:41012
observedproperty=air_pressure_at_sea_level
responseformat=text/csv
eventtime=latest
'''


def parse_unit(prop, dictionary, dt=None):
    '''Do a fuzzy match for `prop` in the dictionary, taking into account unit suffix.'''
    # add the observation's time
    try:
        dt = timezone.parse_datetime(dictionary.get('date_time'))
    except TypeError:
        dt = None

    # 'prop' is a stub of the property's attribute key, so search for matches
    matches = [k for k in dictionary.keys() if prop in k]

    try:
        value = dictionary[matches[0]]
        unit = re.search(r' \(([^)]+)\)', matches[0])

    except IndexError:
        # No matches: fail out
        return None

    # Sometimes we get a list of values (e.g. waves)
    if ';' in value:
        # Ignore empty values
        values = [val for val in value.split(';') if val != '']

        if unit:
            return [Observation(v, unit.group(1), dt) for v in values]
        else:
            return values

    # Sometimes there's no value! Sometimes there's no unit!
    if not value or not unit:
        return value or None

    return Observation(value, unit.group(1), dt)


def _degroup(iterable, propertylist):
    return [{prop: parse_unit(prop, row) for prop in propertylist} for row in iterable]


'''
Response looks like:
station_id,sensor_id,"latitude (degree)","longitude (degree)",date_time,"depth (m)","air_pressure_at_sea_level (hPa)"
urn:ioos:station:wmo:41012,urn:ioos:sensor:wmo:41012::baro1,30.04,-80.55,2014-02-19T12:50:00Z,0.00,1022.1
'''


class Buoy(object):

    '''Wrapper for the NDBC Buoy information mini-API'''

    __dict__ = {}
    params = {
        'request': 'GetObservation',
        'service': 'SOS',
        'version': '1.0.0',
        'responseformat': 'text/csv',
    }

    def __init__(self, bouyid, eventtime=None):
        self.id = bouyid
        self.refresh()

        if eventtime:
            if eventtime.tzinfo:
                eventtime = eventtime.astimezone(utc)
            eventtime = timezone.iso_format(eventtime)
        self.eventtime = eventtime or 'latest'

    def refresh(self):
        self.__dict__ = {
            'lat': None,
            'lon': None,
            'datetime': None,
        }

    def _get(self, observation, as_group=None):
        return self.__dict__.setdefault(observation, self.fetch(observation, as_group))

    def fetch(self, observation, as_group=None):
        params = {
            'offering': 'urn:ioos:station:wmo:{}'.format(self.id),
            'observedproperty': observation,
            'eventtime': self.eventtime
        }
        params.update(self.params)
        request = requests.get(OBS_ENDPOINT, params=params)

        try:
            reader = csv.DictReader(StringIO(request.text))

            if as_group:
                return _degroup(reader, getattr(properties, observation))

            else:
                result = next(reader)

            if 'ows:ExceptionReport' in str(result):
                raise AttributeError(observation)

        except StopIteration:
            raise AttributeError(observation)

        self.__dict__['station_id'] = result.get('station_id')
        self.__dict__['sensor_id'] = result.get('sensor_id')

        try:
            self.__dict__['lon'] = float(result.get('longitude (degree)'))
            self.__dict__['lat'] = float(result.get('latitude (degree)'))

        except TypeError:
            self.__dict__['lon'], self.__dict__['lat'] = None, None

        self.__dict__['depth'] = parse_unit('depth', result)

        return parse_unit(observation, result)

    @property
    def air_pressure_at_sea_level(self):
        return self._get('air_pressure_at_sea_level')

    @property
    def air_temperature(self):
        return self._get('air_temperature')

    @property
    def currents(self):
        try:
            return self._get('currents', as_group=True)
        except IndexError:
            pass

    @property
    def sea_floor_depth_below_sea_surface(self):
        return self._get('sea_floor_depth_below_sea_surface')

    @property
    def sea_water_electrical_conductivity(self):
        return self._get('sea_water_electrical_conductivity')

    @property
    def sea_water_salinity(self):
        return self._get('sea_water_salinity')

    @property
    def sea_water_temperature(self):
        return self._get('sea_water_temperature')

    @property
    def waves(self):
        try:
            return self._get('waves', as_group=True)[0]
        except IndexError:
            pass

    @property
    def winds(self):
        try:
            return self._get('winds', as_group=True)[0]
        except IndexError:
            pass

    @property
    def image_url(self):
        return '{0}?station={id}'.format(CAM_ENDPOINT, id=self.id)

    def _write_img(self, handle):
        i = requests.get(CAM_ENDPOINT, params={'station': self.id})
        for chunk in i.iter_content():
            handle.write(chunk)

    @property
    def image(self):
        output = BytesIO()
        self._write_img(output)
        output.seek(0)

        return output

    def save_image(self, filename):
        with open(filename, 'wb') as f:
            self._write_img(f)

    @property
    def coords(self):
        return self.__dict__.get('lat'), self.__dict__.get('lon')

    @property
    def depth(self):
        return self.__dict__.get('depth')


class Observation(float):

    def __init__(self, value, unit, datetime=None):
        self.value = value
        self._unit = unit
        self._datetime = datetime

    def __new__(cls, value, *args):
        return float.__new__(cls, value)

    @property
    def unit(self):
        return self._unit

    @property
    def datetime(self):
        return self._datetime

    def __repr__(self):
        return "Observation({}, '{}')".format(self.__float__(), self.unit)

    def __str__(self):
        return "{} {}".format(self.__float__(), self.unit)



# -*- coding: utf-8 -*-
# Copyright 2014-17 Neil Freeman
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from setuptools import setup

try:
    readme = open('README.rst').read()
except IOError:
    try:
        readme = open('README.md').read()
    except IOError:
        readme = ''

with open('buoyant/__init__.py') as i:
    version = next(r for r in i.readlines() if '__version__' in r).split('=')[1].strip('"\' \n')

setup(
    name='buoyant',

    version=version,

    description="Wrapper for the NOAA National Data Buoy Center",

    long_description=readme,

    url='https://github.com/fitnr/buoyant',

    author='Neil Freeman',

    author_email='contact@fakeisthenewreal.org',

    license='GPL',

    packages=['buoyant'],

    install_requires=[
        'requests >=2.4.1, <3.0',
        'pytz>=2015.6, <2018'
    ],

    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'Intended Audience :: Science/Research',
    ],

    test_suite='tests',
)



# -*- coding: utf-8 -*-
# Copyright 2014-17 Neil Freeman
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# lat, lon, datetime are assigned separately.
general = [
    'air_pressure_at_sea_level',
    'air_temperature',
    'currents',
    'sea_floor_depth_below_sea_surface',
    'sea_water_electrical_conductivity',
    'sea_water_salinity',
    'sea_water_temperature',
    'waves',
    'winds',
]

currents = [
    'bin',  # (count)
    'depth',  # (m)
    'direction_of_sea_water_velocity',  # (degree)
    'sea_water_speed',  # (c/s)
    'upward_sea_water_velocity',  # (c/s)
    'error_velocity',  # (c/s)
    'platform_orientation',  # (degree)
    'platform_pitch_angle',  # (degree)
    'platform_roll_angle',  # (degree)
    'sea_water_temperature',  # (C)
    'pct_good_3_beam',  # (%)
    'pct_good_4_beam',  # (%)
    'pct_rejected',  # (%)
    'pct_bad',  # (%)
    'echo_intensity_beam1',  # (count)
    'echo_intensity_beam2',  # (count)
    'echo_intensity_beam3',  # (count)
    'echo_intensity_beam4',  # (count)
    'correlation_magnitude_beam1',  # (count)
    'correlation_magnitude_beam2',  # (count)
    'correlation_magnitude_beam3',  # (count)
    'correlation_magnitude_beam4',  # (count)
    'quality_flags',
]

waves = [
    "sea_surface_wave_significant_height",  # (m)
    "sea_surface_wave_peak_period",  # (s)
    "sea_surface_wave_mean_period",  # (s)
    "sea_surface_swell_wave_significant_height",  # (m)
    "sea_surface_swell_wave_period",  # (s)
    "sea_surface_wind_wave_significant_height",  # (m)
    "sea_surface_wind_wave_period",  # (s)
    "sea_water_temperature",  # (c)
    "sea_surface_wave_to_direction",  # (degree)
    "sea_surface_swell_wave_to_direction",  # (degree)
    "sea_surface_wind_wave_to_direction",  # (degree)
    "number_of_frequencies",  # (count)
    "center_frequencies",  # (Hz)
    "bandwidths",  # (Hz)
    "spectral_energy",  # (m**2/Hz)
    "mean_wave_direction",  # (degree)
    "principal_wave_direction",  # (degree)
    "polar_coordinate_r1",  # (1)
    "polar_coordinate_r2",  # (1)
    "calculation_method",
    "sampling_rate",  # (Hz)
]

winds = [
    "wind_from_direction",  # (degree)
    "wind_speed",  # (m/s)
    "wind_speed_of_gust",  # (m/s)
    "upward_air_velocity",  # (m/s)
]
